
Unit AlphaLoaderUnit;   // version 0.1   12.Oct.2005  using graphics32 1.8 beta

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ExtDlgs, Buttons, MMSystem, JPEG, Gr32_System,
  Gr32, GR32_Image, GR32_Layers, GR32_Filters, GR32_LowLevel, Gr32_PNG;

function SavePictureFile (bitmap: tBitmap32; filename: string;
                          transparent: boolean = false): boolean;

type
  TAlphaLoaderForm = class(TForm)
    OpenPictureDialog: TOpenPictureDialog;
    SavePictureDialog: TSavePictureDialog;
    Panel1: TPanel;
      LabelInfo: TLabel;
      Bevel1: TBevel;
    LabelRGBImage: TLabel;
      ImgRGB: TImgView32;
      ButtonLoad: TButton;
      LabelTransparentColor: TLabel;
      transColor: TEdit;
      pixelColor: TEdit;
    LabelAlphaImage: TLabel;
      ImgAlpha: TImgView32;
      LabelTolerance: TLabel;
      ScrollBarTolerance: TScrollBar;
    GroupBox1: TGroupBox;
      SpeedButtonScalePlus: TSpeedButton;
      SpeedButtonScaleMinus: TSpeedButton;
      ButtonFitScale: TButton;
      ButtonResetScale: TButton;
    ButtonSave: TButton;
    ButtonOK: TButton;
    ButtonCancel: TButton;
    LabelRT: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ImgViewPaintStage(img: TImgView32);
    procedure ImgPaintStage(Sender: TObject; Buffer: TBitmap32;
      StageNum: Cardinal);
    procedure CalcTransparentBitmap;
    procedure ButtonLoadClick(Sender: TObject);
    procedure ScrollBarToleranceChange(Sender: TObject);
    procedure SpeedButtonScalePlusClick(Sender: TObject);
    procedure SpeedButtonScaleMinusClick(Sender: TObject);
    procedure ButtonFitScaleClick(Sender: TObject);
    procedure ButtonResetScaleClick(Sender: TObject);
    procedure ButtonSaveClick(Sender: TObject);

    procedure ImgRGBMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure ImgRGBMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer; Layer: TCustomLayer);
    procedure ImgAlphaMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer; Layer: TCustomLayer);
    procedure ImgMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    function ImgRGBColor (X, Y: Integer): TColor32;
  private
    { Private declarations }
    picFilename: string;
    transCol32: tColor32;
  public
    { Public declarations }
  end;

var
  AlphaLoaderForm: TAlphaLoaderForm;

implementation

{$R *.DFM}

//--- picture file routines ---

//*********************************************************
// save bitmap32 as JPEG picture file.
// input:      bitmap       save this bitmap as picture file
//             filename     picture file name (*.jpg)
//             quality      picture quality in % (100=best)
//             progEncoding =true: progressive encoding
// result:     boolean      =true: picture saved
//---------------------------------------------------------
function SaveBitmap32ToJPG (bitmap: tBitmap32;
                            filename: string;
                            quality: integer = 95;
                            progEncoding: boolean = false): boolean;
var    bm: tBitmap;
       jpg: TJPEGImage;
begin
  result := false;
  bm := Graphics.TBitmap.Create;
  try
    jpg := TJpegImage.Create;
    try
      bm.Assign (bitmap);              // convert data into bitmap
      jpg.Assign (bm);
      jpg.CompressionQuality := quality;
      jpg.Compress;    // ???
      jpg.ProgressiveEncoding := progEncoding;
      jpg.SaveToFile (filename);
    finally
      result := true;
      jpg.Free;
      bm.Free;
    end;
  except
  end;
end;

//*********************************************************
// Save bitmap32 as picture file. With this routine you can also save
// '.JPG' and '.PNG' picture files with transparent colors !!!
// Replace GR32.TBitmap32.SaveToFile with this call.
// input:      bitmap      save this bitmap as picture file
//             filename    picture file name (*.bmp, *.jpg, *.png)
//             transparent =true: save transparent background
// result:     boolean     =true: picture saved
//---------------------------------------------------------
function SavePictureFile (bitmap: tBitmap32; filename: string;
                          transparent: boolean = false): boolean;
var    ext: string;
begin
  result := false;
  if (filename = '') then exit;      // ===>
  if bitmap.Empty then exit;         // ===>
  ext := UpperCase (ExtractFileExt (filename));
  if ext = '.PNG'
  //--- save PNG file
  then SaveBitmap32ToPNG (bitmap, transparent, clBlack32, filename)
  else if (ext = '.JPG') or (ext = '.JPEG')
  //--- save JPG file
  then SaveBitmap32ToJPG (bitmap, filename)
  //--- save picture file with windows standard extension
  else bitmap.SaveToFile (filename);
  result := true;
end;

//*********************************************************
//
// procedure SetTransparentBackground (src, dest: tBitmap32;  bgColor: tColor32;
//                               maxDist: integer = 100;  minOpaque: byte = 32);
//
// Description
//  Set every pixel of the destination full transparent, if the source pixel
//  is equal to the given background color.
//
//  The transparency for every pixel will be interpolated, if the color distance
//  is less then the given maximum tolerance distance. The default value for
//  the optional maximum tolerance distance ist 100.
//  Values over 250 are possible, but make sense only in some special cases.
//
//  Because a nearly complete transparent pixel makes no sense, the level of
//  minimum opaqueness can be set optionally. The default value of the minimal
//  opaqueness is 32, the maximum value is 255 (not recommendable).
//  Note that the rgb values of full transparent pixels are cleared completely.
//
//  See also: GR32_Filters.pas, AlphaToGrayscale, IntensityToAlpha, ChromaKey
//
//  input:   src, dest     source and destination bitmap
//           bgColor       background color (complete transparent)
//           maxDist       optional max. distance to all opaque pixel colors
//           minOpaque     optional minimal opaqueness
//                         if pixel opaqueness <= minOpaque -> pixel invisible
//---------------------------------------------------------
procedure SetTransparentBackground (src, dest: tBitmap32;
                                    bgColor: tColor32;
                                    maxDist: integer = 100;
                                    minOpaque: byte = 32);
const  invBlack: TColor32Entry = (a:0);
var    ni, dist: integer;
       bgc: TColor32Entry;        // background color
       pix: PColor32Entry;        // ARGB pixel
begin
  if dest <> src then dest.Assign(src);     // copy bitmap
  maxDist := sqr(maxDist);
  bgc := TColor32Entry (bgColor);
  pix := PColor32Entry (dest.PixelPtr[0, 0]);
  for ni := 0 to (dest.Height * dest.Width - 1)
  do begin
    dist := sqr(pix.r - bgc.r)    // calculate color distance
          + sqr(pix.g - bgc.g)
          + sqr(pix.b - bgc.b);
    if dist < maxDist             // make pixel transparent ?
    then begin                    // change alpha value
      pix.a := MulDiv (dist, 255, maxDist);
      if pix.a < minOpaque        // optimize pixel opaqueness ?
      then pix^ := invBlack;      // set pixel complete invisible
    end;
    inc(pix);                     // proceed to the next pixel
  end;
end;

//---------------------------------------------------------
// calculate transparent bitmap
//---------------------------------------------------------
procedure TAlphaLoaderForm.CalcTransparentBitmap;
begin
  GlobalPerfTimer.Start;
  SetTransparentBackground (ImgRGB.bitmap, ImgAlpha.bitmap
                           ,transCol32, ScrollBarTolerance.Position);
  // !!! delete comment to display runtime in milliseconnds
  labelrt.Caption := 'runtime=' +  GlobalPerfTimer.ReadMilliseconds + ' ms';
end;

procedure TAlphaLoaderForm.FormCreate(Sender: TObject);
begin
  // create chess like background
  with   ImgRGB.PaintStages[0]^ do if Stage = PST_CLEAR_BACKGND then Stage := PST_CUSTOM;
  with ImgAlpha.PaintStages[0]^ do if Stage = PST_CLEAR_BACKGND then Stage := PST_CUSTOM;
end;

// repaint RGB image
procedure TAlphaLoaderForm.FormActivate(Sender: TObject);
begin
  Paint;
end;

// draw chess board texture
procedure TAlphaLoaderForm.ImgViewPaintStage(img: TImgView32);
const
  Colors: array [0..1] of TColor32 = ($FFFFFFFF, $FF99AAAA);
var
  W, I, J, Parity: Integer;
  Line1, Line2: TArrayOfColor32; // a buffer for a couple of scanlines
begin
  with img.Buffer do
  begin
    W := Width;
    SetLength(Line1, W);
    SetLength(Line2, W);
    for I := 0 to W - 1 do
    begin
      Parity := I shr 3 and $1;
      Line1[I] := Colors[Parity];
      Line2[I] := Colors[1 - Parity];
    end;
    for J := 0 to Height - 1 do
    begin
      Parity := J shr 3 and $1;
      if Boolean(Parity) then MoveLongword(Line1[0], ScanLine[J]^, W)
      else MoveLongword(Line2[0], ScanLine[J]^, W);
    end;
  end;
end;

// draw chess board texture of both RGB and Alpha image
// because of Stage = PST_CLEAR_BACKGND
procedure TAlphaLoaderForm.ImgPaintStage(Sender: TObject;
  Buffer: TBitmap32; StageNum: Cardinal);
begin
  ImgViewPaintStage (ImgRGB);
  ImgViewPaintStage (ImgAlpha);
end;

// load new picture file
procedure TAlphaLoaderForm.ButtonLoadClick(Sender: TObject);
var   ext: string;   transparent: boolean;
begin
  with OpenPictureDialog
  do if Execute then begin

    ext := ExtractFileExt (filename);
    if ext = '.png'
    then LoadPNGintoBitmap32 (ImgRGB.Bitmap, filename, transparent)
    else ImgRGB.Bitmap.LoadFromFile(FileName);

    picFilename := filename;
    transCol32 := ImgRGB.bitmap.Pixel[0,0];
    transColor.Color := WinColor (transCol32);
    ImgRGB.Scale := (ImgRGB.Height - 20) / ImgRGB.Bitmap.Height;
    ImgAlpha.Scale := ImgRGB.Scale;
    CalcTransparentBitmap;
    ButtonSave.Enabled := true;
  end;
end;

procedure TAlphaLoaderForm.ScrollBarToleranceChange(Sender: TObject);
begin
  LabelTolerance.Caption := '  tolerance ' + IntToStr (ScrollBarTolerance.Position);
  CalcTransparentBitmap;
end;

procedure TAlphaLoaderForm.SpeedButtonScalePlusClick(Sender: TObject);
begin
  ImgRGB.Scale := ImgRGB.Scale * 1.5;
  ImgAlpha.Scale := ImgRGB.Scale;
end;

procedure TAlphaLoaderForm.SpeedButtonScaleMinusClick(Sender: TObject);
begin
  ImgRGB.Scale := ImgRGB.Scale / 1.5;
  ImgAlpha.Scale := ImgRGB.Scale;
end;

procedure TAlphaLoaderForm.ButtonFitScaleClick(Sender: TObject);
begin
  ImgRGB.Scale := 1;
  if ImgRGB.Bitmap.Height > 10
  then ImgRGB.Scale := (ImgRGB.Height - 20) / ImgRGB.Bitmap.Height;
  ImgAlpha.Scale := ImgRGB.Scale;
end;

procedure TAlphaLoaderForm.ButtonResetScaleClick(Sender: TObject);
begin
  ImgRGB.Scale := 1;
  ImgAlpha.Scale := 1;
end;

// return pixel color from bitmap of given mouse position
function TAlphaLoaderForm.ImgRGBColor (X, Y: Integer): TColor32;
var   sc: double;    xp,yp: integer;
      p: TPoint;
begin
  result := 0;               // return invisible black
  if ImgRGB.Bitmap.Empty then exit;
  sc := ImgRGB.Scale;
  xp := round(x/sc);
  yp := round(y/sc);
  if (xp < 0) or (yp < 0) then exit;
  p := Point(X, Y);
  p := ImgRGB.ControlToBitmap (Point(X, Y));
  result := ImgRGB.Bitmap.PixelS[p.x, p.y];
end;

// set mouse cursor color
procedure TAlphaLoaderForm.ImgRGBMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
begin
  pixelColor.Color := WinColor (ImgRGBColor (x,y));
  ImgAlpha.SetFocus;   // must be set to get mouse wheel events !!!
end;

// select transparent color
procedure TAlphaLoaderForm.ImgRGBMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer;
  Layer: TCustomLayer);
begin
  transCol32 := ImgRGBColor (x,y);
  transColor.Color := WinColor(transCol32);
  if [ssShift] = shift then
  ImgRGB.Bitmap.Assign (ImgAlpha.Bitmap);  // copy ImgAlpha to ImgRGB
  CalcTransparentBitmap;
end;

procedure TAlphaLoaderForm.ImgAlphaMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
begin
  ImgAlpha.SetFocus;   // must be set to get mouse wheel events !!!
end;

procedure TAlphaLoaderForm.ImgMouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
begin
  if wheelDelta > 0
  then ImgRGB.Scale := ImgRGB.Scale * 1.1
  else ImgRGB.Scale := ImgRGB.Scale / 1.1;
  ImgAlpha.Scale := ImgRGB.Scale;
end;

procedure TAlphaLoaderForm.ButtonSaveClick(Sender: TObject);
var   savefilename: string;
begin
  // save transparent PNG picture file
  savefilename := ChangeFileExt (picFilename, '.png');
  SavePictureDialog.FileName := savefilename;
  if SavePictureDialog.Execute
  then SavePictureFile (ImgAlpha.Bitmap, SavePictureDialog.FileName, true);
end;

end.
